"""
        Gelman et al. Bayesian Data Analysis, 3rd ed, p.177
"""
import numpy as np
import stan
import arviz as az

stan_code = """
    data {
        int N;
        array[N] real y;
        array[N] real x;
    }
    parameters {
        //    A noninformative prior distribution, BDA 3, p.165.
        real a;
        real b;
        real log_sgm;
    }
    transformed parameters {
        real sgm;
        sgm = exp(log_sgm);
    }
    model {
        for (i in 1:N) {
            y[i] ~ normal(a + b*x[i], sgm);
        }
    }
    generated quantities {
        array[N] real probs;    //   for calculation of WAIC by Formulas
        array[N] real log_lik;  //   for arviz.waic and loo 
        for (i in 1:N) {
            log_lik[i] = normal_lpdf(y[i] | a + b*x[i], sgm);
            probs[i] = exp(log_lik[i]);
        }
    }
"""

xy = [
[44.6, 2.4],
[57.76, 2.89],
[49.91, 0.85],
[61.34, 4.21],
[49.6, 3.02],
[61.79, 3.62],
[48.95, 1.08],
[44.7, -0.39],
[59.17, 3.86],
[53.94, 2.27],
[46.55, 0.38],
[54.74, 1.04],
[50.27, 2.36],
[51.24, 1.72],
[46.32, 0.1]]
x = np.array(xy).T[1]
y = np.array(xy).T[0]

Data = {'N':len(y), 'y':y, 'x':x}
sm = stan.build(stan_code, data=Data)  
fit = sm.sample() 

data_f = fit.to_frame()
p_list = []
for i in range(len(y)):
    p_list.append(data_f[f'probs.{i+1}'])
p_y_theta = np.array(p_list)
#
#       Calculation of WAIC
#
lppd = np.sum(np.log(np.mean(p_y_theta, axis = 1)))
pWAIC2 = np.sum(np.var(np.log(p_y_theta), axis = 1))
print('\nWAIC =', -2 * (lppd - pWAIC2))
#
#       For arviz.waic and arviz.loo
#
i_data = az.from_pystan(posterior=fit, posterior_model=sm)
vWAIC = az.waic(i_data, scale='deviance')
print('\narviz.WAIC =', vWAIC.elpd_waic)

print('\nVariances =\n', np.var(np.log(p_y_theta), axis = 1, ddof = 1))

vLOOCV = az.loo(i_data, scale='deviance')
print('\nPSIS-LOO-CV =', vLOOCV.elpd_loo)

print('\nWAIC:\n', vWAIC)
print('\nPSIS-LOO-CV:\n', vLOOCV)
